@@ -9,7 +9,7 @@
  * See the file LICENSE for information on usage and redistribution.	
  * ----------------------------------------------------------------------------- */
 
-static char cvsroot[] = "$Header: /cvs/projects/SWIG/Source/Swig/naming.c,v 1.4.4.10 2001/11/29 05:53:56 beazley Exp $";
+char cvsroot_naming_c[] = "$Header: /cvsroot/SWIG/Source/Swig/naming.c,v 1.8 2003/09/11 20:26:57 beazley Exp $";
 
 #include "swig.h"
 #include <ctype.h>
@@ -25,54 +25,26 @@ static Hash *naming_hash = 0;
  * ----------------------------------------------------------------------------- */
 
 void
-Swig_name_register(String_or_char *method, String_or_char *format) {
+Swig_name_register(const String_or_char *method, const String_or_char *format) {
   if (!naming_hash) naming_hash = NewHash();
   Setattr(naming_hash,method,format);
 }
 
-/* -----------------------------------------------------------------------------
- * Swig_name_mangle()
- *
- * Converts all of the non-identifier characters of a string to underscores.
- * ----------------------------------------------------------------------------- */
-
-String *
-Swig_name_mangle(String_or_char *s) {
-  String *r = NewString("");
-  char  *c;
-  Append(r,s);
-  c = Char(r);
-  while (*c) {
-    if (!isalnum(*c)) *c = '_';
-    c++;
+void
+Swig_name_unregister(const String_or_char *method) {
+  if (naming_hash) {
+    Delattr(naming_hash,method);
   }
-  return r;
 }
 
-/* -----------------------------------------------------------------------------
- * Swig_name_wrapper()
- *
- * Returns the name of a wrapper function.
- * ----------------------------------------------------------------------------- */
-
-String *
-Swig_name_wrapper(String_or_char *fname) {
-  String *r;
-  String *f;
-  char *c;
-  int  special = 0;
-  r = NewString("");
-  if (!naming_hash) naming_hash = NewHash();
-  f = Getattr(naming_hash,"wrapper");
-  if (!f) {
-    Append(r,"_wrap_%f");
-  } else {
-    Append(r,f);
-  }
-  Replace(r,"%f",fname, DOH_REPLACE_ANY);
+static int name_mangle(String *r) {
+  char  *c;
+  int    special;
+  special = 0;
+  Replaceall(r,"::","_");
   c = Char(r);
   while (*c) {
-    if (!(isalnum(*c) || (*c == '_'))) {
+    if (!isalnum((int) *c) && (*c != '_')) {
       special = 1;
       switch(*c) {
       case '+':
@@ -132,6 +104,9 @@ Swig_name_wrapper(String_or_char *fname) {
       case '.':
 	*c = 'f';
 	break;
+      case '?':
+	*c = 'q';
+	break;
       default:
 	*c = '_';
 	break;
@@ -139,10 +114,44 @@ Swig_name_wrapper(String_or_char *fname) {
     }
     c++;
   }
-  if (special) {
-    Append(r,"___");
+  if (special) Append(r,"___");
+  return special;
+}
+
+/* -----------------------------------------------------------------------------
+ * Swig_name_mangle()
+ *
+ * Converts all of the non-identifier characters of a string to underscores.
+ * ----------------------------------------------------------------------------- */
+
+String *
+Swig_name_mangle(const String_or_char *s) {
+  String *r = NewString(s);
+  name_mangle(r);
+  return r;
+}
+
+/* -----------------------------------------------------------------------------
+ * Swig_name_wrapper()
+ *
+ * Returns the name of a wrapper function.
+ * ----------------------------------------------------------------------------- */
+
+String *
+Swig_name_wrapper(const String_or_char *fname) {
+  String *r;
+  String *f;
+
+  r = NewString("");
+  if (!naming_hash) naming_hash = NewHash();
+  f = Getattr(naming_hash,"wrapper");
+  if (!f) {
+    Append(r,"_wrap_%f");
+  } else {
+    Append(r,f);
   }
-  /*  Replace(r,":","_", DOH_REPLACE_ANY); */
+  Replace(r,"%f",fname, DOH_REPLACE_ANY);
+  name_mangle(r);
   return r;
 }
 
@@ -154,11 +163,13 @@ Swig_name_wrapper(String_or_char *fname) {
  * ----------------------------------------------------------------------------- */
 
 String *
-Swig_name_member(String_or_char *classname, String_or_char *mname) {
+Swig_name_member(const String_or_char *classname, const String_or_char *mname) {
   String *r;
   String *f;
-  char   *cname, *c;
+  String *rclassname;
+  char   *cname;
 
+  rclassname = SwigType_namestr(classname);
   r = NewString("");
   if (!naming_hash) naming_hash = NewHash();
   f = Getattr(naming_hash,"member");
@@ -167,11 +178,16 @@ Swig_name_member(String_or_char *classname, String_or_char *mname) {
   } else {
     Append(r,f);
   }
-  cname = Char(classname);
-  c = strchr(cname, ' ');
-  if (c) cname = c+1;
+  cname = Char(rclassname);
+  if ((strncmp(cname,"struct ", 7) == 0) ||
+      ((strncmp(cname,"class ", 6) == 0)) ||
+      ((strncmp(cname,"union ", 6) == 0))) {
+    cname = strchr(cname, ' ')+1;
+  }
   Replace(r,"%c",cname, DOH_REPLACE_ANY);
   Replace(r,"%m",mname, DOH_REPLACE_ANY);
+  /*  name_mangle(r);*/
+  Delete(rclassname);
   return r;
 }
 
@@ -182,7 +198,7 @@ Swig_name_member(String_or_char *classname, String_or_char *mname) {
  * ----------------------------------------------------------------------------- */
 
 String *
-Swig_name_get(String_or_char *vname) {
+Swig_name_get(const String_or_char *vname) {
   String *r;
   String *f;
 
@@ -206,7 +222,7 @@ Swig_name_get(String_or_char *vname) {
  * ----------------------------------------------------------------------------- */
 
 String *
-Swig_name_set(String_or_char *vname) {
+Swig_name_set(const String_or_char *vname) {
   String *r;
   String *f;
 
@@ -230,10 +246,13 @@ Swig_name_set(String_or_char *vname) {
  * ----------------------------------------------------------------------------- */
 
 String *
-Swig_name_construct(String_or_char *classname) {
+Swig_name_construct(const String_or_char *classname) {
   String *r;
   String *f;
-  char *cname, *c;
+  String *rclassname;
+  char *cname;
+
+  rclassname = SwigType_namestr(classname);
   r = NewString("");
   if (!naming_hash) naming_hash = NewHash();
   f = Getattr(naming_hash,"construct");
@@ -243,14 +262,52 @@ Swig_name_construct(String_or_char *classname) {
     Append(r,f);
   }
 
-  cname = Char(classname);
-  c = strchr(cname, ' ');
-  if (c) cname = c+1;
-
+  cname = Char(rclassname);
+  if ((strncmp(cname,"struct ", 7) == 0) ||
+      ((strncmp(cname,"class ", 6) == 0)) ||
+      ((strncmp(cname,"union ", 6) == 0))) {
+    cname = strchr(cname, ' ')+1;
+  }
   Replace(r,"%c",cname, DOH_REPLACE_ANY);
+  Delete(rclassname);
+  return r;
+}
+
+
+/* -----------------------------------------------------------------------------
+ * Swig_name_copyconstructor()
+ *
+ * Returns the name of the accessor function used to copy an object.
+ * ----------------------------------------------------------------------------- */
+
+String *
+Swig_name_copyconstructor(const String_or_char *classname) {
+  String *r;
+  String *f;
+  String *rclassname;
+  char *cname;
+
+  rclassname = SwigType_namestr(classname);
+  r = NewString("");
+  if (!naming_hash) naming_hash = NewHash();
+  f = Getattr(naming_hash,"construct");
+  if (!f) {
+    Append(r,"copy_%c");
+  } else {
+    Append(r,f);
+  }
+
+  cname = Char(rclassname);
+  if ((strncmp(cname,"struct ", 7) == 0) ||
+      ((strncmp(cname,"class ", 6) == 0)) ||
+      ((strncmp(cname,"union ", 6) == 0))) {
+    cname = strchr(cname, ' ')+1;
+  }
+
+  Replace(r,"%c",cname, DOH_REPLACE_ANY);
+  Delete(rclassname);
   return r;
 }
-  
 
 /* -----------------------------------------------------------------------------
  * Swig_name_destroy()
@@ -258,10 +315,12 @@ Swig_name_construct(String_or_char *classname) {
  * Returns the name of the accessor function used to destroy an object.
  * ----------------------------------------------------------------------------- */
 
-String *Swig_name_destroy(String_or_char *classname) {
+String *Swig_name_destroy(const String_or_char *classname) {
   String *r;
   String *f;
-  char *cname, *c;
+  String *rclassname;
+  char *cname;
+  rclassname = SwigType_namestr(classname);
   r = NewString("");
   if (!naming_hash) naming_hash = NewHash();
   f = Getattr(naming_hash,"destroy");
@@ -271,14 +330,51 @@ String *Swig_name_destroy(String_or_char *classname) {
     Append(r,f);
   }
 
-  cname = Char(classname);
-  c = strchr(cname, ' ');
-  if (c) cname = c+1;
-
+  cname = Char(rclassname);
+  if ((strncmp(cname,"struct ", 7) == 0) ||
+      ((strncmp(cname,"class ", 6) == 0)) ||
+      ((strncmp(cname,"union ", 6) == 0))) {
+    cname = strchr(cname, ' ')+1;
+  }
   Replace(r,"%c",cname, DOH_REPLACE_ANY);
+  Delete(rclassname);
   return r;
 }
 
+
+/* -----------------------------------------------------------------------------
+ * Swig_name_disown()
+ *
+ * Returns the name of the accessor function used to disown an object.
+ * ----------------------------------------------------------------------------- */
+
+String *Swig_name_disown(const String_or_char *classname) {
+  String *r;
+  String *f;
+  String *rclassname;
+  char *cname;
+  rclassname = SwigType_namestr(classname);
+  r = NewString("");
+  if (!naming_hash) naming_hash = NewHash();
+  f = Getattr(naming_hash,"disown");
+  if (!f) {
+    Append(r,"disown_%c");
+  } else {
+    Append(r,f);
+  }
+
+  cname = Char(rclassname);
+  if ((strncmp(cname,"struct ", 7) == 0) ||
+      ((strncmp(cname,"class ", 6) == 0)) ||
+      ((strncmp(cname,"union ", 6) == 0))) {
+    cname = strchr(cname, ' ')+1;
+  }
+  Replace(r,"%c",cname, DOH_REPLACE_ANY);
+  Delete(rclassname);
+  return r;
+}
+
+
 /* -----------------------------------------------------------------------------
  * Swig_name_object_set()
  *
@@ -288,6 +384,8 @@ String *Swig_name_destroy(String_or_char *classname) {
 void 
 Swig_name_object_set(Hash *namehash, String *name, SwigType *decl, DOH *object) {
   DOH *n;
+
+  /*  Printf(stdout,"name:  '%s', '%s'\n", name, decl);*/
   n = Getattr(namehash,name);
   if (!n) {
     n = NewHash();
@@ -329,10 +427,12 @@ Swig_name_object_get(Hash *namehash, String *prefix, String *name, SwigType *dec
 
   if (!namehash) return 0;
 
-  if ((decl) && (SwigType_isqualifier(decl))) {
+  /* DB: This removed to more tightly control feature/name matching */
+  /*  if ((decl) && (SwigType_isqualifier(decl))) {
     ncdecl = strchr(Char(decl),'.');
     ncdecl++;
   }
+  */
 
   /* Perform a class-based lookup (if class prefix supplied) */
   if (prefix) {
@@ -380,7 +480,7 @@ Swig_name_object_get(Hash *namehash, String *prefix, String *name, SwigType *dec
 
 void
 Swig_name_object_inherit(Hash *namehash, String *base, String *derived) {
-  String *key;
+  Iterator ki;
   String *bprefix;
   String *dprefix;
   char *cbprefix;
@@ -392,23 +492,23 @@ Swig_name_object_inherit(Hash *namehash, String *base, String *derived) {
   dprefix = NewStringf("%s::",derived);
   cbprefix = Char(bprefix);
   plen = strlen(cbprefix);
-  for (key = Firstkey(namehash); key; key = Nextkey(namehash)) {
-    char *k = Char(key);
+  for (ki = First(namehash); ki.key; ki = Next(ki)) {
+    char *k = Char(ki.key);
     if (strncmp(k,cbprefix,plen) == 0) {
       Hash *n, *newh;
-      String *nkey, *okey;
-      
+      String *nkey;
+      Iterator oi;
+
       nkey = NewStringf("%s%s",dprefix,k+plen);
-      n = Getattr(namehash,key);
+      n = ki.item;
       newh = Getattr(namehash,nkey);
       if (!newh) {
 	newh = NewHash();
 	Setattr(namehash,nkey,newh);
       }
-      for (okey = Firstkey(n); okey; okey = Nextkey(n)) {
-	String *ovalue = Getattr(n,okey);
-	if (!Getattr(newh,okey)) {
-	  Setattr(newh,okey,Copy(ovalue));
+      for (oi = First(n); oi.key; oi = Next(oi)) {
+	if (!Getattr(newh,oi.key)) {
+	  Setattr(newh,oi.key,Copy(oi.item));
 	}
       }
     }
@@ -422,14 +522,14 @@ Swig_name_object_inherit(Hash *namehash, String *base, String *derived) {
  * ----------------------------------------------------------------------------- */
 
 static void merge_features(Hash *features, Node *n) {
-  String *fs;
-  String *key;
+  Iterator ki;
+
   if (!features) return;
-  for (key = Firstkey(features); key; key = Nextkey(features)) {
-    if (Getattr(n,key)) {
+  for (ki = First(features); ki.key; ki = Next(ki)) {
+    if (Getattr(n,ki.key)) {
       continue;
     }
-    Setattr(n,key,Copy(Getattr(features,key)));
+    Setattr(n,ki.key,Copy(ki.item));
   }
 }
 
@@ -439,7 +539,7 @@ Swig_features_get(Hash *features, String *prefix, String *name, SwigType *decl,
   DOH    *rn = 0;
   Hash   *n;
   char   *ncdecl = 0;
-
+  
   if (!features) return;
 
   if ((decl) && (SwigType_isqualifier(decl))) {
@@ -447,10 +547,47 @@ Swig_features_get(Hash *features, String *prefix, String *name, SwigType *decl,
     ncdecl++;
   }
 
-  /* Perform a class-based lookup (if class prefix supplied) */
-  if (prefix) {
-    if (Len(prefix)) {
-      tname = NewStringf("%s::%s",prefix,name);
+  if (name) {
+    /* Perform a class-based lookup (if class prefix supplied) */
+    if (prefix) {
+      if (Len(prefix)) {
+	tname = NewStringf("%s::%s",prefix,name);
+	n = Getattr(features,tname);
+	rn = get_object(n,decl);
+	merge_features(rn,node);
+	if (ncdecl) {
+	  rn = get_object(n,ncdecl);
+	  merge_features(rn,node);
+	}
+	rn = get_object(n,0);
+	merge_features(rn,node);
+	Delete(tname);
+      }
+      /* A wildcard-based class lookup */
+      tname = NewStringf("*::%s",name);
+      n = Getattr(features,tname);
+      rn = get_object(n,decl);
+      merge_features(rn,node);
+      if (ncdecl) {
+	rn = get_object(n,ncdecl);
+	merge_features(rn,node);
+      }
+      rn = get_object(n,0);
+      merge_features(rn,node);
+      Delete(tname);
+      
+      /* A class-generic feature */
+      if (Len(prefix)) {
+	tname = NewStringf("%s::",prefix);
+	n = Getattr(features,tname);
+	rn = get_object(n,0);
+	merge_features(rn,node);
+	Delete(tname);
+      }
+      
+    } else {
+      /* Lookup in the global namespace only */
+      tname = NewStringf("::%s",name);
       n = Getattr(features,tname);
       rn = get_object(n,decl);
       merge_features(rn,node);
@@ -462,9 +599,8 @@ Swig_features_get(Hash *features, String *prefix, String *name, SwigType *decl,
       merge_features(rn,node);
       Delete(tname);
     }
-    /* A wildcard-based class lookup */
-    tname = NewStringf("*::%s",name);
-    n = Getattr(features,tname);
+    /* Catch-all */
+    n = Getattr(features,name);
     rn = get_object(n,decl);
     merge_features(rn,node);
     if (ncdecl) {
@@ -473,42 +609,7 @@ Swig_features_get(Hash *features, String *prefix, String *name, SwigType *decl,
     }
     rn = get_object(n,0);
     merge_features(rn,node);
-    Delete(tname);
-
-    /* A class-generic feature */
-    if (Len(prefix)) {
-      tname = NewStringf("%s::",prefix);
-      n = Getattr(features,tname);
-      rn = get_object(n,0);
-      merge_features(rn,node);
-      Delete(tname);
-    }
-
-  } else {
-    /* Lookup in the global namespace only */
-    tname = NewStringf("::%s",name);
-    n = Getattr(features,tname);
-    rn = get_object(n,decl);
-    merge_features(rn,node);
-    if (ncdecl) {
-      rn = get_object(n,ncdecl);
-      merge_features(rn,node);
-    }
-    rn = get_object(n,0);
-    merge_features(rn,node);
-    Delete(tname);
   }
-  /* Catch-all */
-  n = Getattr(features,name);
-  rn = get_object(n,decl);
-  merge_features(rn,node);
-  if (ncdecl) {
-    rn = get_object(n,ncdecl);
-    merge_features(rn,node);
-  }
-  rn = get_object(n,0);
-  merge_features(rn,node);
-
   /* Global features */
   n = Getattr(features,"");
   rn = get_object(n,0);
